package facade;

import DTO.MapBoardDTO;
import DTO.PlayerMapBoardDTO;
import DTO.PlayerDTO;
import DataMapper.DataMapperCalls;
import com.google.common.collect.MapMaker;
import entity.MapBoard;
import entity.MapValues;
import entity.Player;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.EntityManager;
import javax.persistence.Query;
import org.json.JSONArray;
import org.json.JSONObject;
import org.json.JSONTokener;
import utils.PuSelector;

public class Facade {

    private ConcurrentMap<String, Player> playerCache = new MapMaker().concurrencyLevel(4).makeMap();
    private ConcurrentMap<String, MapBoard> mapBoardCache = new MapMaker().concurrencyLevel(4).makeMap();
    private ConcurrentMap<String, ConcurrentMap<String, MapValues>> playerRelatedMapValuesCache = new MapMaker().concurrencyLevel(6).makeMap();
    private final long EXPIRE_TIME_IN_SECONDS = TimeUnit.SECONDS.convert(30, TimeUnit.MINUTES);
    private Instant elapsedTime;
    private int initSetup = 0;

    public Facade() {
        elapsedTime = Instant.now();
        initSetup = 1;
    }

    public List<MapBoardDTO> getSpecificMap(String mapname, int stage) {
        List<MapBoardDTO> returnlist = new ArrayList();
        String mapRetrieved = mapname + "S" + stage;
        MapBoard mapBoard = mapBoardCache.getOrDefault(mapRetrieved, null);
        if (mapBoard == null){
            mapRetrieved = mapname;
            mapBoard = mapBoardCache.get(mapRetrieved);
        }
        List<MapValues> mapvalues = mapBoard.getMapvalues();
        for (MapValues mapvalue : mapvalues) {
            returnlist.add(new MapBoardDTO(mapvalue, playerCache.get(mapvalue.getPlayerSteamID()).getName()));
        }
        return returnlist;
    }

    public List<PlayerMapBoardDTO> getPlayerMaps(String steamid) {
        try {
            ConcurrentMap<String, MapValues> playerMaps = playerRelatedMapValuesCache.get(steamid);
            List<PlayerMapBoardDTO> playerMapBoardDtos = new ArrayList();
            for (Entry<String, MapValues> entries : playerMaps.entrySet()) {
                PlayerMapBoardDTO mapBoardDto = new PlayerMapBoardDTO(steamid);
                mapBoardDto.addMapValues(entries.getValue(), entries.getKey());
                playerMapBoardDtos.add(mapBoardDto);
            }
            return playerMapBoardDtos;
        } catch (Exception ex) {
            return null;
        }
    }

    public PlayerDTO getPlayer(String steamid) {
        EntityManager em = PuSelector.getEntityManagerFactory("pu").createEntityManager();
        try {
            Query query = em.createQuery("SELECT a FROM Player AS a WHERE a.steamID LIKE :steamID")
                    .setParameter("steamID", steamid);
            Player player = (Player) query.getSingleResult();
            return new PlayerDTO(player);
        } finally {
            em.close();
        }
    }

    public Collection<PlayerDTO> getleaderBoard() {
        EntityManager em = PuSelector.getEntityManagerFactory("pu").createEntityManager();
        try {
            Query query = em.createQuery("SELECT a FROM Player AS a ORDER BY A.rank ASC ");
            List<Player> players = (List<Player>) query.getResultList();
            Collection<PlayerDTO> playersDTO = new ArrayList();
            for (Player player : players) {
                playersDTO.add(new PlayerDTO(player));
            }
            return playersDTO;
        } finally {
            em.close();
        }
    }

    private void addNewPlayerValuesToCache(Collection<Player> keyCollection) {
        for (Player player : keyCollection) {
            playerCache.put(player.getSteamID(), player);
        }
    }

    private String convertCommunityIdToSteamId(long communityId) {
        long steamId1 = communityId % 2;
        long steamId2 = communityId - 76561197960265728L;
        steamId2 = (steamId2 - steamId1) / 2;
        return "STEAM_0:" + steamId1 + ":" + steamId2;
    }

    private void retrieveAvatarFull(String avatarurl) {
        //https://steamcommunity.com/profiles/76561198517905428   some nosteamers have no information available on steams rest-endpoints
        try {
            URL url = new URL(avatarurl);
            //System.out.println("avatarurl: " + avatarurl + "\n");
            JSONTokener tokener = new JSONTokener(url.openStream());
            JSONObject root = new JSONObject(tokener);
            JSONObject jsonObject = root.getJSONObject("response");
            JSONArray jsonArray = jsonObject.getJSONArray("players");
            int incrementer = 0;
            for (; incrementer < jsonArray.length();) {
                System.out.println("incrementer: " + incrementer + "\n");
                try {
                    String steamID64 = jsonArray.getJSONObject(incrementer).getString("steamid");
                    String avatar = jsonArray.getJSONObject(incrementer).getString("avatarfull");
                    String steamID = convertCommunityIdToSteamId(Long.valueOf(steamID64));
                    playerCache.get(steamID).setAvatar(avatar);
                } catch (Exception ex) {
                    System.out.println("ex: " + ex.getLocalizedMessage() + "\n");
                }
                incrementer++;
            }
        } catch (MalformedURLException ex) {
            Logger.getLogger(Facade.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(Facade.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private void updateAvatarsToCache() {
        //http://api.steampowered.com/ISteamUser/GetPlayerSummaries/v0002/?key=262F3F0C6B83E0F263C272C3762002F1&steamids=76561198029832363,76561198003907342
        StringBuilder avatarURL = new StringBuilder();
        avatarURL.append("http://api.steampowered.com/ISteamUser/GetPlayerSummaries/v0002/?key=262F3F0C6B83E0F263C272C3762002F1&steamids=");
        int incrementer = 0;
        for (Player player : playerCache.values()) {
            if (incrementer > 0) {
                avatarURL.append(",");
            }
            incrementer++;
            avatarURL.append(Long.toString(convertSteamIdToCommunityId(player.getSteamID())));
            //550 is too large size for url.openStream(), fetch limit is probably about 100 players best
            if (incrementer >= 100 || incrementer >= playerCache.size()) {
                retrieveAvatarFull(avatarURL.toString());
                avatarURL = new StringBuilder();
                avatarURL.append("http://api.steampowered.com/ISteamUser/GetPlayerSummaries/v0002/?key=262F3F0C6B83E0F263C272C3762002F1&steamids=");
                incrementer = 0;
            }
        }
    }

    private void setNonExistingAvatarOnPlayers() {
        for (Player player : playerCache.values()) {
            String avatar = player.getAvatar();
            if (avatar == null) {
                player.setAvatar("https://steamcdn-a.akamaihd.net/steamcommunity/public/images/avatars/fe/fef49e7fa7e1997310d705b2a6158ff8dc1cdfeb_full.jpg");
            }
        }
    }

    public static long convertSteamIdToCommunityId(String steamId) {
        if (steamId.matches("^STEAM_[0-1]:[0-1]:[0-9]+$")) {
            String[] tmpId = steamId.substring(8).split(":");
            return Long.valueOf(tmpId[0]) + Long.valueOf(tmpId[1]) * 2 + 76561197960265728L; //76561197960265728L
        } else if (steamId.matches("^\\[U:[0-1]:[0-9]+\\]+$")) {
            String[] tmpId = steamId.substring(3, steamId.length() - 1).split(":");
            return Long.valueOf(tmpId[0]) + Long.valueOf(tmpId[1]) + 76561197960265728L; //76561197960265727L
        }
        return 0;
    }

    private void addNewMapBoardValuesToCache(Collection<MapBoard> value) {
        for (MapBoard mapboard : value) {
            mapBoardCache.put(mapboard.getMapName(), mapboard);
        }
    }

    private void orderingMapBoardValuesByTime() {
        Collection<MapBoard> mapboards = mapBoardCache.values();
        for (MapBoard mapboard : mapboards) {
            List<MapValues> mapvalues = new ArrayList(mapboard.getMapvalues());
            mapvalues.sort(Comparator.comparing(MapValues::getTime));
            mapBoardCache.get(mapboard.getMapName()).setMapvalues(mapvalues);
        }
    }

    private void setPositionAndMapPoints() {
        for (MapBoard mapboard : mapBoardCache.values()) {
            List<MapValues> mapvalues = mapboard.getMapvalues();
            int playerCount = mapvalues.size();
            int position = 1;
            for (MapValues mapvalue : mapvalues) {
                mapvalue.setPosition(position);
                float percentageSpreader = 0;
                percentageSpreader = playerCount - position;
                mapvalue.setMapPoints((int) Math.floor((percentageSpreader / playerCount) * 100));
                position++;
            }
        }
    }

    private void setPlayerPoints() {
        for (MapBoard mapboard : mapBoardCache.values()) {
            List<MapValues> mapvalues = mapboard.getMapvalues();
            for (MapValues mapvalue : mapvalues) {
                String steamID = mapvalue.getPlayerSteamID();
                int mapPoints = mapvalue.getMapPoints();
                System.out.println("\nmapPoints: " + mapPoints + "\n");
                playerCache.get(steamID).addPlayerPoints(mapPoints);
            }
        }
    }

    private void updatePlayerRanks() {
        List<Player> playerList = new ArrayList(playerCache.values());
        playerList.sort(Comparator.comparing(Player::getPlayerPoints).reversed());
        int rank = 1;
        for (Player player : playerList) {
            playerCache.get(player.getSteamID()).setRank(rank);
            rank++;
        }
    }

    public void setTimesAsociatedToPeople() {
        Collection<MapBoard> mapboards = mapBoardCache.values();
        for (MapBoard mapboard : mapboards) {
            List<MapValues> mapvalues = mapboard.getMapvalues();
            for (MapValues mapvalue : mapvalues) {
                ConcurrentMap<String, MapValues> playerMapValueCache = playerRelatedMapValuesCache.getOrDefault(mapvalue.getPlayerSteamID(), null);
                if (playerMapValueCache == null) {
                    ConcurrentMap<String, MapValues> localMap = new MapMaker().concurrencyLevel(4).makeMap();
                    localMap.put(mapboard.getMapName(), mapvalue);
                    playerRelatedMapValuesCache.put(mapvalue.getPlayerSteamID(), localMap);
                } else {
                    playerMapValueCache.put(mapboard.getMapName(), mapvalue);
                    playerRelatedMapValuesCache.put(mapvalue.getPlayerSteamID(), playerMapValueCache);
                }
            }
        }
    }

    public void checkTimeElapse() {
        Instant current = Instant.now();
        long timeElapsed = Duration.between(elapsedTime, current).toMillis();
        System.out.println("timeElapsed: " + timeElapsed + "\n");
        if (timeElapsed >= EXPIRE_TIME_IN_SECONDS || initSetup > 0) {
            initSetup = 0;
            elapsedTime = Instant.now();
            Entry<Collection<Player>, Collection<MapBoard>> playersMapBoards = DataMapperCalls.getAllValues();
            Collection<MapBoard> mapBoardCollection = playersMapBoards.getValue();
            playerCache.clear();
            mapBoardCache.clear();
            playerRelatedMapValuesCache.clear();
            addNewMapBoardValuesToCache(mapBoardCollection);
            addNewPlayerValuesToCache(playersMapBoards.getKey());
            updateAvatarsToCache();
            setNonExistingAvatarOnPlayers();
            orderingMapBoardValuesByTime();
            setPositionAndMapPoints();
            setPlayerPoints();
            updatePlayerRanks();
            setTimesAsociatedToPeople();
        }
    }

    public List<Player> getAllPlayersFromCache() {
        checkTimeElapse();
        return new ArrayList(playerCache.values());
    }

    public Collection<MapBoard> getAllMapBoardsFromCache() {
        checkTimeElapse();
        return mapBoardCache.values();
    }

}
